#define CORE_PRIVATE

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"

#include <stdio.h>

#include <lve/lve-ctl.h>

#define PREFIX "mod_hostinglimits:"

module MODULE_VAR_EXPORT hostinglimits_module;

//static APR_OPTIONAL_FN_TYPE(vh_info_lookup) *pfn_vh_info_lookup = NULL;

static int lve_available = 0;
struct liblve *lve = NULL;
static uint32_t p_cookie = 0;

// configuration data
typedef struct hostinglimits_module_cfg {
  unsigned int skip;
  uint32_t lve_id;
  array_header *allowed_handlers;
} hostinglimits_module_cfg;

static hostinglimits_module_cfg *hostinglimits_module_dconfig(request_rec *r){
  return (hostinglimits_module_cfg *) ap_get_module_config(r->per_dir_config,&hostinglimits_module);
}

static void *hostinglimits_module_create_dir_config(pool *p, char *dirspec)
{

  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *)ap_pcalloc(p, sizeof(hostinglimits_module_cfg));
  if(!cfg)  {
      ap_log_error(APLOG_MARK, APLOG_ERR, NULL, PREFIX" not enough memory");
      return NULL;
  }
  cfg->skip  = 0;
  cfg->lve_id = 0;
  cfg->allowed_handlers = NULL;
  return (void *) cfg;

}

static void *hostinglimits_module_merge_config(pool* p, void* BASE, void* ADD) {

  hostinglimits_module_cfg* base = BASE;
  hostinglimits_module_cfg* add = ADD;
  hostinglimits_module_cfg* cfg = (hostinglimits_module_cfg *)ap_pcalloc(p, sizeof(hostinglimits_module_cfg));
  cfg->skip = (add->skip)?add->skip:base->skip;
  cfg->lve_id = (add->lve_id)?add->lve_id:base->lve_id;
  cfg->allowed_handlers = (add->allowed_handlers)?add->allowed_handlers:base->allowed_handlers;
  return (void *)cfg;

}

static const char *set_lve_id(cmd_parms *cmd, void *mcfg, const char *lve_id)
{
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  if (lve_id) {
      cfg->lve_id = (uint32_t)atoi(lve_id);
  }
  return NULL;
}

static const char *set_skip(cmd_parms *cmd, void *mcfg, const char *skip){
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  if (skip) {
      cfg->skip = (unsigned int)atoi(skip);
  }
  return NULL;
}

static const char *set_handlers(cmd_parms *cmd, void *mcfg, const char *arg){
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  const char *err = ap_check_cmd_context(cmd, NOT_IN_DIR_LOC_FILE | NOT_IN_LIMIT);
  if (err != NULL) {
    return err;
  }
  if (!cfg->allowed_handlers) {
    cfg->allowed_handlers = ap_make_array(cmd->pool, 2, sizeof(char *));
  }
  *(const char **)ap_push_array(cfg->allowed_handlers) = arg;
  return NULL;
}

static int hostinglimits_module_exit(request_rec *r)
{
   if (!p_cookie) {
       return DECLINED;
   }
   if (lve_available) {
     hostinglimits_module_cfg  *cfg = hostinglimits_module_dconfig(r);

    if(!cfg) {
     return DECLINED;
    }
    int rc=lve_leave(lve, &p_cookie);
    if(!rc) {
     p_cookie = 0;
     return DECLINED;
    }
    ap_log_error(APLOG_MARK, APLOG_ERR, r->server,PREFIX"Error exiting %d", rc);
    //error exiting, lets kill it
    exit(-1);
  }

  return DECLINED;
}

static void hostinglimits_initializer(server_rec *s, pool *p){
  int rc;
  if (lve == NULL) {
     lve = ap_palloc(p, lve_instance_init(NULL));
     rc = lve_instance_init(lve);
     if (rc != 0) {
       lve_available = 0;
     } else {
       lve_available = 1;
       ap_register_cleanup(p, lve, (void*)lve_instance_destroy, ap_null_cleanup);
     }
  }
  //pfn_vh_info_lookup = APR_RETRIEVE_OPTIONAL_FN(vh_info_lookup);
  //i can't find APR_RETRIEVE_OPTIONAL_FN analog for apache 1.3.
  //it comes from apr_utils, will be apr_utils correct work with apache 1.3?
  //i disabled this function
  if (lve_available) {
    ap_log_error(APLOG_MARK, APLOG_NOTICE, s, "LVE mechanism enabled");
    //printf("LVE mechanism enabled\n");
  } else {
    ap_log_error(APLOG_MARK, APLOG_ERR, s, "LVE mechanism disabled, LVE is unavailable");
    //printf("LVE mechanism disabled, LVE is unavailable\n");
  }
}

static int match_handler(array_header *allowed_handlers, const char *handler) {
  int num_names;
  char **names_ptr;
  int accept = 0;
  if (allowed_handlers && handler) {
    names_ptr = (char **)allowed_handlers->elts;
    num_names = allowed_handlers->nelts;
  } else {
    return 0;
  }
  // match handle
  for (; num_names; ++names_ptr, --num_names) {
    if (!strcmp(handler, *names_ptr)) {
        accept=1;
        break;
    }
  }
  return accept;
}

static int enter_lve_fixups(request_rec *r) {
  // get directory config info

  hostinglimits_module_cfg *cfg = hostinglimits_module_dconfig(r);
  if(p_cookie || !cfg) {
    // we are already in lve, or config not initiated - request declined
    return DECLINED;
  }
  if (!lve_available && !cfg->skip) {
    ap_log_error(APLOG_MARK, APLOG_ERR, r->server, PREFIX"LVE is unavailable LVE");
    return HTTP_SERVICE_UNAVAILABLE;
  }
  if (lve_available) {
    //lets find handlers for which HostingLimits should run
    if ((!match_handler(cfg->allowed_handlers, r->handler))&&(!match_handler(cfg->allowed_handlers, r->content_type))) {
        return DECLINED;
    }

    //enter into lve
    uid_t    uid = -1;
    int lve_id = cfg->lve_id;

    uid = r->server->server_uid;
    if (lve_id < 1) {
        lve_id = uid;
    }
    if (lve_id > 0) {
      int rc = lve_enter(lve, lve_id, -1, -1, &p_cookie); //todo add limits
      if(rc) {
        if (errno == EPERM) { //if already inside LVE
           ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
                             PREFIX"Already inside LVE: LVE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s)",
                             lve_id, r->handler, r->hostname, r->uri);
           return DECLINED;
        }
        ap_log_error(APLOG_MARK, APLOG_ERR, r->server, PREFIX"Error on LVE enter %d", errno);
        p_cookie = 0;
        return HTTP_SERVICE_UNAVAILABLE;
      }
    }
  }
  return DECLINED;
}

static command_rec hostinglimits_module_directives[] = {
  {"SkipErrors", set_skip, NULL, RSRC_CONF, TAKE1, "Allow apache to continue even if LVE is unavalable"},
  {"LVEId", set_lve_id, NULL, ACCESS_CONF | RSRC_CONF, TAKE1, "LVE Id"},
  {"AllowedHandlers", set_handlers, NULL, RSRC_CONF, ITERATE,"A list of handlers which will be put in LVE"},
  {NULL}
};


module hostinglimits_module = {
  STANDARD_MODULE_STUFF,
  hostinglimits_initializer,                     /* initializer */
  hostinglimits_module_create_dir_config,  /* dir config creator */
  hostinglimits_module_merge_config,       /* dir merger --- default is to override */
  NULL,   /* server config */
  NULL,  /* merge server config */
  hostinglimits_module_directives,                 /* command table */
  NULL,             /* handlers */
  NULL,    /* filename translation */
  NULL,                     /* check_user_id */
  NULL,                     /* check auth */
  NULL,                     /* check access */
  NULL,         /* type_checker */
  enter_lve_fixups,                     /* fixups */
  hostinglimits_module_exit,        /* logger */
  NULL,                      /* header parser */
  NULL,      /* [3] header parser */
  NULL,         /* process initializer */
  NULL,         /* process exit/cleanup */
  NULL   /* [1] post read_request handling */
};

#undef CORE_PRIVATE
