/**
 *
 * This is the hostinglimits module for apache 2.2.X
 * author Pookhov Sergey (pookhov@gmail.com)
 * author Igor Seletskiy <iseletsk@gmail.com>
 * How to make, install and setup see README file
 *
 */
#include "apr_general.h"
#include "apr_lib.h"
#include "apr_strings.h"
#include "ap_config.h"
#include "util_filter.h"
#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"
#include "http_core.h"
#include <unistd.h>
#include "unixd.h"

#include <lve/lve-ctl.h>

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/stat.h>

#define PREFIX "mod_hostinglimits:"

module AP_MODULE_DECLARE_DATA hostinglimits_module;
static apr_threadkey_t *key;
static apr_thread_once_t *once_object = NULL;
static apr_pool_t *ppool;

// configuration data
typedef struct hostinglimits_module_cfg {
  unsigned int skip;
  uint32_t lve_id;
  apr_array_header_t *allowed_handlers;
} hostinglimits_module_cfg;

static int lve_available = 0;
struct liblve *lve = NULL;


/**
 * This function returns directory config structure for current request
 */
static hostinglimits_module_cfg *hostinglimits_module_dconfig(const request_rec *r) {
  return (hostinglimits_module_cfg *) ap_get_module_config(r->per_dir_config,
                                                           &hostinglimits_module);
}


static void *hostinglimits_module_merge_config(apr_pool_t* p, void* BASE, void* ADD) {
  hostinglimits_module_cfg* base = BASE;
  hostinglimits_module_cfg* add = ADD;
  hostinglimits_module_cfg* cfg = (hostinglimits_module_cfg *)
    apr_pcalloc(p, sizeof(hostinglimits_module_cfg));
  cfg->skip = (add->skip)?add->skip:base->skip;
  cfg->lve_id = (add->lve_id)?add->lve_id:base->lve_id;
  cfg->allowed_handlers = (add->allowed_handlers)?add->allowed_handlers:
    base->allowed_handlers;
  return cfg;
}

static void *hostinglimits_module_create_dir_config(apr_pool_t *p,
                                                    char *dirspec) {
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *)
    apr_pcalloc(p, sizeof(hostinglimits_module_cfg));
  if(!cfg) {
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, NULL,
                   PREFIX" not enough memory");
    return NULL;
  }
  cfg->skip  = 0;
  cfg->lve_id = 0;
  cfg->allowed_handlers = NULL;
  return (void *) cfg;
}


static const char *set_lve_id(cmd_parms *cmd, void *mcfg,
                              const char *lve_id)
{
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  if (lve_id) {
    cfg->lve_id = (uint32_t)apr_atoi64(lve_id);;
  }
  return NULL;
}


static const char *set_handlers(cmd_parms *cmd, void *mcfg, const char *arg)
{
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  const char *err = ap_check_cmd_context(cmd, NOT_IN_DIR_LOC_FILE|NOT_IN_LIMIT);
  if (err != NULL) {
    return err;
  }
  if (!cfg->allowed_handlers) {
    cfg->allowed_handlers = apr_array_make(cmd->pool, 2, sizeof(char *));
  }
  *(const char **)apr_array_push(cfg->allowed_handlers) = arg;
  return NULL;
}
// describe used directives
static command_rec hostinglimits_module_directives[] = {
  AP_INIT_FLAG("SkipErrors", ap_set_flag_slot,  (void *)
               APR_OFFSETOF(hostinglimits_module_cfg, skip),
               RSRC_CONF, "Allow apache to continue even if LVE is unavalable"),
  AP_INIT_TAKE1("LVEId", set_lve_id, NULL, ACCESS_CONF | RSRC_CONF, "LVE Id"),
  AP_INIT_ITERATE("AllowedHandlers", set_handlers, NULL, RSRC_CONF,
                  "A list of handlers which will be put in LVE"),

  {NULL}
};

static int init_post_config(apr_pool_t *pconf, apr_pool_t *plog,
                        apr_pool_t *ptemp, server_rec *s) {
  if (lve_available) {
      ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,
                   PREFIX"LVE mechanism enabled");
  } else {
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
                   PREFIX"LVE mechanism disabled, LVE is unavailable");
  }

  return OK;
}

static int match_handler(apr_array_header_t *allowed_handlers, const char *handler) {
  int num_names;
  char **names_ptr;
  int accept = 0;
  if (allowed_handlers) {
      names_ptr = (char **)allowed_handlers->elts;
      num_names = allowed_handlers->nelts;
    } else {
      return 0;
    }
    // match handle
    for (; num_names; ++names_ptr, --num_names) {
      if (!strcmp(handler, *names_ptr)) {
        accept=1;
        break;
      }
    }
    return accept;
}

/**
 * Where are main functionality in this function.
 * This funcion performs after reading request headers
 * and before other proceccing.
 *
 *
 */
static int enter_lve_handler(request_rec *r) {
  // get directory config info
  uint32_t *p_cookie = NULL;
  apr_status_t st;
  hostinglimits_module_cfg *cfg = hostinglimits_module_dconfig(r);
  if (!lve_available || !cfg) {
    return DECLINED;
  }
  /* decline if cookie is set ( p_cookie != 0) */
  st=apr_threadkey_private_get((void **)&p_cookie, key);
  if(st != APR_SUCCESS || (p_cookie!=NULL && *p_cookie != 0)) {
     return DECLINED;
  }
  //lets find handlers for which HostingLimits should run
  if (!match_handler(cfg->allowed_handlers, r->handler)) {
    return DECLINED;
  }

  //enter into lve
  uid_t    uid = -1;
  int lve_id = cfg->lve_id;

  ap_unix_identity_t * ugid = ap_run_get_suexec_identity(r);
  if (ugid == NULL) {
      uid = -1;
  } else {
      uid = ugid->uid;
  }
  if (lve_id < 1) {
      lve_id = uid;
  }

  if (lve_id > 0) {
    p_cookie = (uint32_t*)apr_pcalloc(r->pool, sizeof(uint32_t));
    if(!p_cookie) {
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
                   PREFIX"Can't allocate memory");
      return HTTP_SERVICE_UNAVAILABLE;
    }
    int rc = lve_enter(lve, lve_id, -1, -1, p_cookie); //todo add limits
    if(rc) {
      if (errno == EPERM) { //if already inside LVE
        ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
                     PREFIX"Already inside LVE: LVE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s)",
                     lve_id, r->handler, r->hostname, r->uri);
        return DECLINED;
      }
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
                   PREFIX"Error on LVE enter %d", errno);
      return HTTP_SERVICE_UNAVAILABLE;
    }
    st = apr_threadkey_private_set(p_cookie, key);

    if (st != APR_SUCCESS){
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
                   PREFIX"Unable to set LVE cookie %d", st);
    }
  }

  return DECLINED;
}


static int leave_lve_handler(request_rec *r) {
  uint32_t *p_cookie;
  apr_status_t st;
  if (!lve_available) {
    return DECLINED;
  }

  st=apr_threadkey_private_get((void **)&p_cookie, key);
  /* if cookie is not initialized or if it is not 0 -- don't try to leave */
  if (st != APR_SUCCESS || p_cookie == NULL || *p_cookie == 0){
    return DECLINED;
  }

  hostinglimits_module_cfg  *cfg = hostinglimits_module_dconfig(r);

  int rc = lve_leave(lve, p_cookie);
  if(!rc) {
    /* reset cookie to 0 */
    *p_cookie = 0;
    st = apr_threadkey_private_set(NULL, key);

    if (st != APR_SUCCESS){
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
                   PREFIX"Unable to reset LVE cookie %d", st);
    }
    return DECLINED;
  }
  ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
               PREFIX"Error exiting %d", rc);
  //error exiting, lets kill it
  exit(-1);
}

static apr_status_t destroy_key_pool(void * dummy)
{
  apr_threadkey_private_delete(key);
  apr_pool_destroy(ppool);
  ppool = 0;
  return APR_SUCCESS;
}


static void hostinglimits_child_init(apr_pool_t *p, server_rec *s){
  if (APR_SUCCESS == apr_pool_create ((&ppool), 0)){
   apr_pool_cleanup_register (p, 0, destroy_key_pool, destroy_key_pool);
   apr_threadkey_private_create(&key, NULL, ppool);
  } else {
    ap_log_error(APLOG_MARK, APLOG_ERR, OK, s,
               PREFIX"Can't create Key pool");
    exit(-1);
  }

}

// function for hook register
static void hostinglimits_module_register_hooks(apr_pool_t *p) {
  int rc;
  if (lve == NULL) {
    lve = apr_palloc(p, lve_instance_init(NULL));
    rc = lve_instance_init(lve);
    if (rc != 0) {
      lve_available = 0;
    } else {
      lve_available = 1;
      // register clean up LVE
      apr_pool_cleanup_register(p, lve, (void*)lve_instance_destroy, apr_pool_cleanup_null);
    }
  }

  static const char * const aszPre[] = { "mod_include.c", "mod_php.c", "mod_cgi.c", NULL };

  ap_hook_post_config(init_post_config, NULL, NULL, APR_HOOK_MIDDLE);
  // hook executes before request prforming
  ap_hook_handler(enter_lve_handler, NULL, aszPre, APR_HOOK_REALLY_FIRST);
  // hook executes after data generation and logging
  ap_hook_log_transaction(leave_lve_handler, NULL, NULL, APR_HOOK_LAST);
  ap_hook_child_init(hostinglimits_child_init,NULL,NULL,APR_HOOK_MIDDLE);

}

/**
 * Describing structure of Apache module
 */
module AP_MODULE_DECLARE_DATA hostinglimits_module = {
  STANDARD20_MODULE_STUFF,
  hostinglimits_module_create_dir_config,   /* create directory config */
  hostinglimits_module_merge_config,        /* merging directory config */
  NULL,                                     /* create server config */
  NULL,                                     /* merging server config */
  hostinglimits_module_directives,          /* mapping configuration directives */
  hostinglimits_module_register_hooks       /* registering hooks */
};
